#include "fish.h"
#include <memory>
#include <cmath>

#if !defined M_PI
#define M_PI 3.14159265358979323846
#endif

using namespace std;

Fish::Fish()
{
  shared_ptr<Vertex> a = make_shared<Vertex>(0, 0, 0);
  shared_ptr<Vertex> b = make_shared<Vertex>(1, -1, 0);
  shared_ptr<Vertex> c = make_shared<Vertex>(1, 0, 1);
  shared_ptr<Vertex> d = make_shared<Vertex>(2, 3, 0);
  shared_ptr<Vertex> e = make_shared<Vertex>(6, 1, 0);

  shared_ptr<Vertex> f = make_shared<Vertex>(7, -1, 0);
  shared_ptr<Vertex> g = make_shared<Vertex>(7, 3, 0);

  shared_ptr<Vertex> h = make_shared<Vertex>(1, 0, -1);

  Face bodyLeft1;
  bodyLeft1.addVertex(a)
    .addVertex(b)
    .addVertex(c);

  Face bodyLeft2;
  bodyLeft2.addVertex(c)
    .addVertex(b)
    .addVertex(e);

  Face bodyLeft3;
  bodyLeft3.addVertex(e)
    .addVertex(d)
    .addVertex(c);

  Face bodyLeft4;
  bodyLeft4.addVertex(c)
    .addVertex(d)
    .addVertex(a);

  Face bodyRight1;
  bodyRight1.addVertex(a)
    .addVertex(h)
    .addVertex(b);

  Face bodyRight2;
  bodyRight2.addVertex(b)
    .addVertex(h)
    .addVertex(e);

  Face bodyRight3;
  bodyRight3.addVertex(e)
    .addVertex(h)
    .addVertex(d);

  Face bodyRight4;
  bodyRight4.addVertex(d)
    .addVertex(h)
    .addVertex(a);

  Face tailLeft;
  tailLeft.addVertex(e)
    .addVertex(f)
    .addVertex(g);

  Face tailRight;
  tailRight.addVertex(e)
    .addVertex(g)
    .addVertex(f);

  shared_ptr<FaceGroup> fishGroup = make_shared<TriangleGroup>();
  fishGroup->
    addFace(bodyLeft1)->
    addFace(bodyLeft2)->
    addFace(bodyLeft3)->
    addFace(bodyLeft4)->
    addFace(bodyRight1)->
    addFace(bodyRight2)->
    addFace(bodyRight3)->
    addFace(bodyRight4)->
    addFace(tailLeft)->
    addFace(tailRight);

  addGroup(fishGroup);
  direction = Vector(-1, 0, 0);
  upside = Vector(0, 1, 0);
  finalize();
}

shared_ptr<Figure> Fish::getFigurePointer()
{
  return shared_from_this();
}

shared_ptr<Fish> Fish::clone() const
{
  shared_ptr<Fish> result = make_shared<Fish>();
  cloneInto(result->getFigurePointer());
  return result;
}

void Fish::applyTransformations()
{
  straightenOut();
  Figure::applyTransformations();
}

void Fish::turnInDirection(const Vector& newDirection, float weight)
{
  if (newDirection == Vector(0, 0, 0))
    return;
  Vector normalizedDirection = newDirection.normal();
  float scalarProduct = direction & normalizedDirection;
  if (scalarProduct > 1)
    scalarProduct = 1;
  else if (scalarProduct < -1)
    scalarProduct = -1;
  float angle = acos(scalarProduct) * (180 / M_PI);
  Vector rotationAxis = direction * normalizedDirection;
  rotate(rotationAxis, weight * (angle / 360), center);
}

void Fish::straightenOut()
{
  if (upside == Vector(0, 1, 0))
    return;
  Vector currentRight = (direction * upside).normal();
  Vector correctedRight = (direction * Vector(0, 1, 0)).normal();
  float dotProduct = currentRight & correctedRight;
  if (dotProduct > 1)
    dotProduct = 1;
  else if (dotProduct < -1)
    dotProduct = -1;

  Vector rotationAxis = currentRight * correctedRight;
  float angle = acos(dotProduct) * (180 / M_PI);

  rotate(rotationAxis, angle, center);

  float angleToY = acos(direction & Vector(0, 1, 0)) * (180 / M_PI);
  if (angleToY < 30)
    rotate(correctedRight, angleToY - 30, center);
  else if (angleToY > 150)
    rotate(correctedRight, angleToY - 150, center);
}

void Fish::turnTowards(const Vector& point, float weight)
{
  turnInDirection(Vector(point - center), weight);
}

void Fish::rotate(const Vector& rotationAxis, float angle, const Vector& center)
{
  Matrix4 plainRotationMatrix = rotationMatrix(rotationAxis.normal(), angle, Vector(0, 0, 0));
  direction = Vector(plainRotationMatrix * direction).normal();
  upside = Vector(plainRotationMatrix * upside).normal();
  Figure::rotate(rotationAxis, angle, center);
}

void Fish::rotate(const Vector& rotationAxis, float angle)
{
  rotate(rotationAxis, angle, Vector(0, 0, 0));
}

void Fish::turnAwayFrom(const Vector& point, float degrees)
{
  turnTowards(point, -degrees);
}

void Fish::moveForward(float amount)
{
  Figure::move(Vector(direction / 10 * amount));
}
