#include "util.h"
#include <cmath>
#include <cfloat>
#include <memory>
#include <exception>
#include "imageloader.h"
using namespace std;

/*
 *
 * Paintable
 */

void Paintable::beforePaint(int shadingType) {}
void Paintable::afterPaint(int shadingType) {}

void Paintable::paint(int shadingType)
{
  beforePaint(shadingType);
  paintMethod(shadingType);
  afterPaint(shadingType);
}
/* end Paintable */

/*
 * Vertex
 */
void Vertex::paintMethod(int shadingType)
{
  if (shadingType == SMOOTH_SHADING)
    glNormal3f(normal.x, normal.y, normal.z);
  glVertex3f(point.x, point.y, point.z);
}
/* end Vertex */

/*
 * Face
 */

void Face::paintMethod(int shadingType)
{
  if (shadingType == FLAT_SHADING)
    glNormal3f(normal.x, normal.y, normal.z);
  Vector texCoords [vertices.size()];
  if (useTexture) {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glPushAttrib(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_TEXTURE_2D);
    if (vertices.size() == 3) {
      texCoords[0] = Vector(0, 0, 0);
      texCoords[1] = Vector(1, 0, 0);
      texCoords[2] = Vector(0.5, 1, 0);
    } else {
      texCoords[0] = Vector(0, 0, 0);
      texCoords[1] = Vector(1, 0, 0);
      texCoords[2] = Vector(1, 1, 0);
      texCoords[3] = Vector(0, 1, 0);
    }
  }
  for (int i = 0; i < vertices.size(); i++) {
    if (useTexture)
      glTexCoord2d(texCoords[i].x, texCoords[i].y);
    vertices[i]->paint(shadingType);
  }
  if (useTexture) {
    glDisable(GL_TEXTURE_2D);
    glPopAttrib();
  }
}

Vector Face::getNormal() const
{
  return normal;
}

const Vector& Face::getBase() const
{
  return vertices[0]->point;
}

Vector Face::closestPointTo(const Vector& point) const
{
  float sb, sn, sd;
  sn = -(normal & Vector(point - vertices[0]->point));
  sd = normal & normal;
  sb = sn / sd;
  return Vector(point + (normal * sb));
}

float Face::distanceTo(const Vector& point) const
{
  return point.distance(closestPointTo(point));
}

void Face::setNormal()
{
  Vector a1 = vertices[1]->point - vertices[0]->point;
  Vector a2 = vertices[2]->point - vertices[0]->point;
  normal = (a1 * a2).normal();
}

Face Face::clone() const
{
  Face result;
  for (auto& vertex : vertices)
    result.addVertex(make_shared<Vertex>(*vertex));
  return result;
}

void Face::sumNormals()
{
  for (auto& vertex : vertices)
    vertex->normal = Vector(vertex->normal + normal);
}

Face& Face::addVertex(shared_ptr<Vertex> vertex)
{
  vertices.push_back(vertex);
  return *this;
}

vector<shared_ptr<Vertex> >& Face::getVertices()
{
  return vertices;
}

/*
 * FaceGroup
 */

void FaceGroup::paintMethod(int shadingType)
{
  for (auto& face : faces)
    face.paint(shadingType);
}

void FaceGroup::afterPaint(int shadingType)
{
  glEnd();
}

void FaceGroup::cloneInto(shared_ptr<FaceGroup> target) const
{
  target->faces = vector<Face>(faces.size());
  for (auto& face : faces)
    target->addFace(face.clone());
}

shared_ptr<FaceGroup> FaceGroup::addFace(Face face)
{
  faces.push_back(face);
  return shared_from_this();
}

vector<Face>& FaceGroup::getFaces()
{
  return faces;
}

const vector<Face>& FaceGroup::getFaces() const
{
  return faces;
}

set<shared_ptr<Vertex> > FaceGroup::getVertices()
{
  set<shared_ptr<Vertex> > vertices;
  for (auto& face : faces) {
    vector<shared_ptr<Vertex> >& verticesForFace = face.getVertices();
    vertices.insert(verticesForFace.begin(), verticesForFace.end());
  }
  return vertices;
}
/* end FaceGroup */

/*
 * TriangleGroup
 */
void TriangleGroup::beforePaint(int shadingType)
{
  glBegin(GL_TRIANGLES);
}

shared_ptr<FaceGroup> TriangleGroup::getFaceGroupPointer()
{
  return shared_from_this();
}

shared_ptr<FaceGroup> TriangleGroup::clone() const
{
  shared_ptr<TriangleGroup> result = make_shared<TriangleGroup>();
  for (auto& face : faces)
    result->addFace(face.clone());

  return result;
}
/* end TriangleGroup */

/*
 * QuadGroup
 */
void QuadGroup::beforePaint(int shadingType)
{
  glBegin(GL_QUADS);
}

shared_ptr<FaceGroup> QuadGroup::clone() const
{
  shared_ptr<QuadGroup> result = make_shared<QuadGroup>();
  for (auto& face : faces)
    result->addFace(face.clone());
  return result;
}

shared_ptr<FaceGroup> QuadGroup::getFaceGroupPointer()
{
  return shared_from_this();
}
/* end QuadGroup */

/*
 * Material
 */

void Material::apply() const
{
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, shininess);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ambient.v);
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, diffuse.v);
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, specular.v);
  glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, emission.v);
}
/*
 * Figure
 */

Figure::Figure()
{
  finalized = false;
}

Figure::Figure(const Figure& other)
{
  finalized = false;
  for (auto& group : other.groups)
    addGroup(group->clone());
  center = other.center;
  boundingRadius = other.boundingRadius;
  material = other.material;
  finalized = other.finalized;
}

void Figure::cloneInto(shared_ptr<Figure> other) const
{
  for (auto& group : groups)
    other->groups.push_back(group->clone());

  other->transformations = transformations;
  other->center = center;
  other->boundingRadius = boundingRadius;
  other->material = material;
  other->finalized = finalized;
}

void Figure::finalize()
{
  setNormals();
  applyTransformations();
  finalized = true;
}

set<shared_ptr<Vertex> > Figure::getVertices()
{
  set<shared_ptr<Vertex> > result;
  for (auto& group : groups) {
    set<shared_ptr<Vertex> > verticesForGroup = group->getVertices();
    result.insert(verticesForGroup.begin(), verticesForGroup.end());
  }
  return result;
}

Figure& Figure::addGroup(shared_ptr<FaceGroup> group)
{
  if (finalized)
    throw exception();
  groups.push_back(group);
  return *this;
}

bool Figure::collidesWith(const Figure& other) const
{
  GLfloat distance = center.distance(other.center);
  return distance <= boundingRadius || distance <= other.boundingRadius;
}

bool Figure::insideBounds(const Figure& bounds, float padding) const
{
  bool result = true;
  for (auto& face : bounds.getFaces()) {
    if (face.distanceTo(center) <= boundingRadius + padding) {
      result = false;
      break;
    }
  }
  return result;
}

bool Figure::insideBounds(const Figure& bounds) const
{
  return insideBounds(bounds, 0);
}

std::vector<Face> Figure::getFaces()
{
  vector<Face> faces;
  for (auto& group : groups) {
    vector<Face>& facesForGroup = group->getFaces();
    faces.insert(faces.end(), facesForGroup.begin(), facesForGroup.end());
  }
  return faces;
}

const std::vector<Face> Figure::getFaces() const
{
  vector<Face> faces;
  for (auto& group : groups) {
    vector<Face>& facesForGroup = group->getFaces();
    faces.insert(faces.end(), facesForGroup.begin(), facesForGroup.end());
  }
  return faces;
}

void Figure::resetNormals(set<shared_ptr<Vertex> >& vertices)
{
  for (auto& vertex : vertices) {
    Vector zero = Vector(0, 0, 0);
    vertex->normal = zero;
  }
}

void Figure::normalizeNormals(set<shared_ptr<Vertex> >& vertices)
{
  for (auto& vertex : vertices)
    vertex->normal.normalize();
}

void Figure::paintMethod(int shadingType)
{
  vector<shared_ptr<FaceGroup> >::iterator groupPtr;
  for (auto& group : groups)
    group->paint(shadingType);
}

void Figure::setNormals()
{
  set<shared_ptr<Vertex> > vertices = getVertices();
  resetNormals(vertices);
  for (auto& group : groups) {
    vector<Face>& facesForGroup = group->getFaces();
    for (auto& face : facesForGroup) {
      face.setNormal();
      face.sumNormals();
    }
  }
  normalizeNormals(vertices);
}

void Figure::applyTransformations()
{
  if (transformations == Matrix4(1))
    return;

  set<shared_ptr<Vertex>> vertices = getVertices();
  for (auto& vertex : vertices) {
    vertex->point = Vector(transformations * vertex->point);
    vertex->normal = Vector(normalTransformations * vertex->normal).normal();
  }
  vector<Face> faces = getFaces();
  for (auto& group : groups) {
    vector<Face>& facesForGroup = group->getFaces();
    for (auto& face : facesForGroup)
      face.setNormal();
  }
  transformations = Matrix4(1);
  calculateBoundingSphere();
}

void Figure::calculateBoundingSphere()
{
  vector<Vector> vectors;
  set<shared_ptr<Vertex> > vertices = getVertices();
  for (auto& vertex : vertices)
    vectors.push_back(vertex->point);

  center = Vector::average(vectors);

  GLfloat maxDistance = 0;
  for (auto& vertex : vertices) {
    GLfloat distance = center.distance(vertex->point);
    if (distance > maxDistance)
      maxDistance = distance;
  }

  boundingRadius = maxDistance;
}

void Figure::beforePaint(int shadingType)
{
  material.apply();
  applyTransformations();
}
